//@line 36 "/builds/buildbot/linux_repack/build/releases/l10n-mozilla-1.9.1/ca/calendar/sunbird-l10n.js"

//@line 38 "/builds/buildbot/linux_repack/build/releases/l10n-mozilla-1.9.1/ca/calendar/sunbird-l10n.js"

//@line 41 "/builds/buildbot/linux_repack/build/releases/l10n-mozilla-1.9.1/ca/calendar/sunbird-l10n.js"
pref("calendar.week.start", 1);

//@line 44 "/builds/buildbot/linux_repack/build/releases/l10n-mozilla-1.9.1/ca/calendar/sunbird-l10n.js"
pref("calendar.week.d0sundaysoff", true);
pref("calendar.week.d1mondaysoff", false);
pref("calendar.week.d2tuesdaysoff", false);
pref("calendar.week.d3wednesdaysoff", false);
pref("calendar.week.d4thursdaysoff", false);
pref("calendar.week.d5fridaysoff", false);
pref("calendar.week.d6saturdaysoff", true);

pref("general.useragent.locale", "ca");

//@line 55 "/builds/buildbot/linux_repack/build/releases/l10n-mozilla-1.9.1/ca/calendar/sunbird-l10n.js"
pref("calendar.categories.names", "Aniversari,Naixament,Negoci,Trucades,Clients,Competència,Clients,Preferits,Seguiments,Regals,Vacances,Idees,Assumptes,Cita,Misceŀlània,Personal,Projectes,Dia de festa,Estat,Proveïdors,Viatge,Descans");
